using System;
using Zcu.Mve.Core;
using System.Xml;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Simple line between two points.
	/// </summary>
	public struct Line : ICell
	{
		/// <summary>
		/// Index of first vertex of line.
		/// </summary>
		public int V1;
		/// <summary>
		/// Index of second vertex of line.
		/// </summary>
		public int V2;

		/// <summary>
		/// Simple parametric constructor.
		/// </summary>
		/// <param name="V1">Index of first vertex.</param>
		/// <param name="V2">Index of second vertex.</param>
		public Line(int V1, int V2)
		{
			this.V1 = V1;
			this.V2 = V2;
		} // Line()

		/// <summary>
		/// Returns array of two vertices of the line.
		/// Change of returned array do not cause change inside line object.
		/// </summary>
		/// <remarks>This access is not efficient. Use V1, V2 instead.</remarks>
		[ObsoleteAttribute("This access is not efficient. Use V1, V2 instead.", false)]
		public int [] Points
		{
			get
			{
				return new int [2] {V1, V2};
			}
		} // PointArray

		/// <summary>
		/// Return number of points of cell.
		/// </summary>
		public int PointCount
		{
			get
			{
				return 2;
			}
		} // PointCount

		/// <summary>
		/// Returns index of vertex.
		/// </summary>
		/// <param name="index">0 or 1 are only acceptable vaules.</param>
		/// <returns>Index of vertex</returns>
		/// <remarks>This access is not efficient. Use V1, V2 instead.</remarks>
		[ObsoleteAttribute("This access is not efficient. Use V1, V2 instead.", false)]
		public int GetPoint(int index)
		{
			switch (index)
			{
				case 0:
					return V1;
				case 1:
					return V2;
				default:
					throw new MveException("Index out of range. Line have only have vertices.");
			}
		} // GetPoint()

		/// <summary>
		/// Get/Set index of line vertex
		/// </summary>
		/// <remarks>This access is not efficient. Use V1, V2 instead.</remarks>
		[ObsoleteAttribute("This access is not efficient. Use V1, V2 instead.", false)]
		public int this[int index]
		{
			get
			{
				switch (index)
				{
					case 0:
						return V1;
					case 1:
						return V2;
					default:
						throw new MveException("Index out of range. Line have only two vertices.");
				}					
			}
			set
			{
				switch (index)
				{
					case 0:
						V1 = value;
						break;
					case 1:
						V2 = value;
						break;
					default:
						throw new MveException("Index out of range. Line have only two vertices.");
				}						
			}
		} // this[int]

		/// <summary>
		/// Read data from XML file.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		public void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			string [] vertices = xmlTextReader.ReadString().Trim().Split(' ', '\t');

			if (vertices.Length != 2)
				throw new MveException("Incorect number of vertices in line.");

			this.V1 = Int32.Parse(vertices[0].Trim());
			this.V2 = Int32.Parse(vertices[1].Trim());
		} // ReadData()

		/// <summary>
		/// Write data to XML file.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		public void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			xmlTextWriter.WriteString(V1 + " " + V2);
		} // WriteData()

		/// <summary>
		/// Deep copy of curent object
		/// </summary>
		/// <returns>Deep copy (same data, different memory place)</returns>
		public IDataObject DeepCopy()
		{
			return new Line(this.V1, this.V2);
		} // DeepCopy()

		/// <summary>
		/// Line with negative index is considered inconsistent.
		/// If two identical indices are detected warning is given.
		/// </summary>
		/// <returns>True if the Line is consistent.</returns>
		public bool CheckConsistence()
		{
			bool isOK = true;
			if (this.V1 < 0 || this.V2 < 0) 
			{
				Console.WriteLine("Error! " + this.GetType().Name + " has negative index.");
				isOK = false;
			}
			if (this.V1 == this.V2)
				Console.WriteLine("Warning! " + this.GetType().Name + " Line deformation: points have identical index.");
			return isOK;
		} // CheckConsistence()

		/// <summary>
		/// Line information
		/// </summary>
		/// <returns>String that contain name of this object and indices of curent object.</returns>
		override public string ToString()
		{
			return String.Format("Line: [{0}, {1}]", V1, V2);
		} // ToString()	

		/// <summary>
		/// Returns array of used assembly names.
		/// </summary>
		public string [] LibNames
		{
			get
			{
				return new string [] {Globals.GetLibraryName(this.GetType())};
			}
		}

	} // class Line
} // namespace
