using System;
using System.Xml;
using System.Collections;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Heterogenous array of DataObjects. 
	/// </summary>
	public class NonUniformDataArray : AbstractDataArray
	{
    private Hashtable elemTypes = new Hashtable();

    /// <summary>
    /// For XmlLoader use only. Use parametric construcotr instead.
    /// </summary>
    public NonUniformDataArray()
    {

    }

    /// <summary>
    /// Creates fixed size, one dimensional array of DataObjects
    /// </summary>
    /// <param name="type">Type of object (only this type and its childrens are accepted). Must be inherited from Mve.Core.DataObject</param>
    /// <param name="count">Number of array elements</param>
		public NonUniformDataArray(Type type, int count) : base(type, count)
		{
		}
  

    /// <summary>
    /// Recursive deep copy of curent array.
    /// </summary>
    /// <returns></returns>
    public override DataObject DeepCopy()
    {
      UniformDataArray array = new UniformDataArray(type, Count);
      for (int i = 0; i < Count; i++)
      {
        array[i] = this[i].DeepCopy();
      }
      return null;
    }
  

    /// <summary>
    /// Prida prvek do pole a upravi seznam obsazenejch datovejch typu.
    /// </summary>
    /// <param name="index"></param>
    /// <param name="element"></param>
    private void UpdateElemTypeTable(int index, DataObject element)
    {
      if (list[index] != null)
      {
        int cnt = (int) this.elemTypes[list[index].GetType()];
        if (cnt == 1)
          elemTypes.Remove(list[index].GetType());
        else
          elemTypes[list[index].GetType()] = cnt - 1;
      }

      if (element == null)
        return;

      if (elemTypes.ContainsKey(element.GetType()))
      {
        int cnt = (int) this.elemTypes[element.GetType()];
        this.elemTypes[element.GetType()] = cnt + 1;
      }
      else
      {
        elemTypes.Add(element.GetType(), 1);
      }
    }


    /// <summary>
    /// Returns list of types that are contained in array.
    /// </summary>
    /// <returns></returns>
    public ICollection GetListOfTypes()
    {
      return elemTypes.Keys;
    }


    /// <summary>
    /// Returns number of types contained in array.
    /// </summary>
    /// <param name="type"></param>
    /// <returns></returns>
    public int GetCountOfType(Type type)
    {
      if (elemTypes.ContainsKey(type))
        return (int) elemTypes[type];
      else
        return 0;
    }

    /// <summary>
    /// Get/Set element of data array.
    /// </summary>
    public DataObject this [int index]
    {
      get
      {
        return list[index];
      }
      set
      {
        if (type.IsInstanceOfType(value))
        {
          if (InvokeElementSetEvent(index, value))
          {
            UpdateElemTypeTable(index, value);
            list[index] = value;
          }
          else
          {
            throw new MveException("Assignement of " + value + " is forbidden by registered callback.");
          }
        }
        else
        {
          throw new MveException("Curent array can't contain instace of " + value.GetType().FullName + ". Only " + type.FullName + " is allowed.");
        }
      }
    }


    private void ReadArray(System.Xml.XmlTextReader xmlTextReader)
    {
      int count = 0;
      int i = 0;
/*
      if (typeName != null)
        throw new MveException("array can be used only once.");
*/
      count = Int32.Parse(xmlTextReader.GetAttribute("count").Trim());
  
      this.list = new DataObject [count];

      while(xmlTextReader.Read())
      {
        if (xmlTextReader.NodeType == XmlNodeType.Element)
        {
          switch(xmlTextReader.Name)
          {
            case "elem":
              if (i < count)
              {
                string emtpyAttr = xmlTextReader.GetAttribute("empty");
                if (emtpyAttr != null && emtpyAttr == "true")
                {
                  list[i] = null;
                }
                else
                {
                  string typeName = xmlTextReader.GetAttribute("type");
                  type = Globals.FindType(this.LibNames, typeName);
                  DataObject obj = (DataObject) Activator.CreateInstance(type);
                  obj.ReadData(xmlTextReader);
                  list[i] = obj;
                }
                i++;
              }
              else
              {
                throw new MveException("Incorrect number of elements in DataArray");
              }
              break;
          }
        }
      }
    }


    /// <summary>
    /// Read array from XML file.
    /// </summary>
    /// <param name="xmlTextReader"></param>
    public override void ReadData(System.Xml.XmlTextReader xmlTextReader)
    {
      while(xmlTextReader.Read())
      {
        if (xmlTextReader.NodeType == XmlNodeType.Element)
          switch(xmlTextReader.Name)
          {
            case "array":     // nacteni pole
              ReadArray(xmlTextReader);
              break;
            default:
              break;
          } // switch
      } // while
    }
  

    /// <summary>
    /// Write whole content of DataArray to XML file.
    /// </summary>
    /// <param name="xmlTextWriter"></param>
    public override void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
    {
      xmlTextWriter.WriteStartElement("array");
      xmlTextWriter.WriteAttributeString("count", this.Count.ToString());
      // xmlTextWriter.WriteAttributeString("elemtype", type.FullName);

      for(int i = 0; i < this.Count; i++)
      {
        DataObject data = list[i];
        if (data != null)
        {
          xmlTextWriter.WriteStartElement("elem");
          xmlTextWriter.WriteAttributeString("type", data.GetType().FullName);
          data.WriteData(xmlTextWriter);
          xmlTextWriter.WriteEndElement();
        }
        else
        {
          xmlTextWriter.WriteStartElement("elem");
          xmlTextWriter.WriteAttributeString("empty", "true");          
          xmlTextWriter.WriteEndElement();
        }
      }       
      xmlTextWriter.WriteEndElement();
    }

    public override ArrayList LibNames
    {
      get
      {
        for (int i = 0; i < this.Count; i++)
        {
          if (list[i] != null)
            list[i].AddLibName(list[i]);
        }
        return base.LibNames;
      }
    }

    /// <summary>
    /// Creates NonUniformDataArray from ArrayList. 
    /// Elements are not copied. Reference only is passed to new array.
    /// </summary>
    /// <param name="type">Base type of nonuniform data array.</param>
    /// <param name="list">List to be converted (copied).</param>
    /// <returns>New instance of NonUniformDataArray.</returns>
    public static NonUniformDataArray FromArrayList(Type type, ArrayList list)
    {
      NonUniformDataArray ret = new NonUniformDataArray(type, list.Count);

      for (int i = 0; i < list.Count; i++)
        ret[i] = (DataObject) list[i];

      return ret;
    }
  }
}
