using System;
using System.Globalization;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Class represents 2D normal vector defined by two elements - Nx and Ny.
	/// </summary>
	public struct Normal2D : INormal
	{
		/// <summary>
		/// x-normal value.
		/// </summary>
		public double Nx;
		/// <summary>
		/// y-normal value.
		/// </summary>
		public double Ny;

		/// <summary>
		/// The toleration for equality operator.
		/// </summary>
		public static double Epsilon = Globals.EpsilonDoubleImplicit;

		/// <summary>
		/// Creates a structure and fill in given values
		/// </summary>
		/// <param name="nx">x-normal value.</param>
		/// <param name="ny">y-normal value.</param>
		public Normal2D(double nx, double ny)
		{
			this.Nx = nx;
			this.Ny = ny;
		} // Normal2D()

		/// <summary>
		/// Read data from Xml file to current structure.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		public void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
			string[] split = xmlTextReader.ReadString().Trim().Split(' ');
			this.Nx = Double.Parse(split[0], nfi);
			this.Ny = Double.Parse(split[1], nfi);
		} // ReadData()

		/// <summary>
		/// Write current data to Xml file.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		public void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
			xmlTextWriter.WriteString(this.Nx.ToString(nfi) + " " + this.Ny.ToString(nfi));
		} // WriteData()

		/// <summary>
		/// Deep copy of data object and its subelements.
		/// </summary>
		/// <returns>Same data, different memory space.</returns>
		public IDataObject DeepCopy()
		{
			return new Normal2D(this.Nx, this.Ny);
		} // DeepCopy()

		/// <summary>
		/// Method check whether absolute value equals 1 (+- NormalND.EPSILON). 
		/// If no, normal is normalized.
		/// </summary>
		public void Normalize()
		{
			double absValue = NormalND.AbsValue(this.Nx, this.Ny);
			this.Nx = this.Nx / absValue;
			this.Ny = this.Ny / absValue;
		} // Normalize();

		/// <summary>
		/// Check whether normal absolute value equals 1 +- NormalND.EPSILON.
		/// </summary>
		/// <returns>Returns true if normal absolute value equals 1 +- NormalND.EPSILON.</returns>
		public bool CheckConsistence()
		{
			double absValue = NormalND.AbsValue(this.Nx, this.Ny);
			if (absValue < 1 - NormalND.EPSILON || absValue > 1 + NormalND.EPSILON)
				return false;
			else
				return true;
		} // CheckConsistence()

		/// <summary>
		/// Get number of elements of normal.
		/// </summary>
		public int Length
		{
			get
			{
				return 2;
			}
		}

		/// <summary>
		/// Get/Set element of normal.
		/// </summary>
		public double this [int index]
		{
			get
			{
				switch(index)
				{
					case 0:
						return this.Nx;
					case 1:
						return this.Ny;
					default:
						throw new IndexOutOfRangeException(this.GetType().FullName + ": Dimension exceeded.");
				}
			}
			set
			{
				switch(index)
				{
					case 0:
						this.Nx = value;
						break;
					case 1:
						this.Ny = value;
						break;
					default:
						throw new IndexOutOfRangeException(this.GetType().FullName + ": Dimension exceeded.");
				}
			}
		}

		/// <summary>
		/// Returns distance from beginning of coordinate system.
		/// </summary>
		/// <returns></returns>
		public double Abs()
		{
			return Math.Sqrt(this.Nx * this.Nx + this.Ny * this.Ny);
		} // Abs()

		/// <summary>
		/// Returns array of assembly name.
		/// </summary>
		public string [] LibNames
		{
			get
			{
				return new string [] {Globals.GetLibraryName(this.GetType())};
			}
		}

		// ===== Methods of the class Object ==============================================
		#region Overrided methods of the class Object

		/// <summary>
		/// Returns a String that represents the current Object.
		/// </summary>
		/// <returns>String that represents the current Object.</returns>
		override public string ToString()
		{
			return "Normal2D: [" + this.Nx + "; " + this.Ny + "]";
		} // ToString()

		/// <summary>
		/// Determines whether the specified Objects are equal.
		/// </summary>
		/// <param name="a">First Object.</param>
		/// <param name="b">Second Object.</param>
		/// <returns>True if objects are equal. Otherwise, false.</returns>
		public static new bool Equals(object a, object b)
		{
			if (a.GetType() != typeof(Normal2D) || b.GetType() != typeof(Normal2D))
				return false;
		      
			return Math.Abs(((Normal2D)a).Nx - ((Normal2D)b).Nx) <= Normal2D.Epsilon
				&& Math.Abs(((Normal2D)a).Ny - ((Normal2D)b).Ny) <= Normal2D.Epsilon;
		}
		
		/// <summary>
		/// Determines whether the specified Object is equal to the current Object.
		/// </summary>
		/// <param name="obj">Specified Object to compare with the current Object.</param>
		/// <returns>True if the specified Object is equal to the current Object. Otherwise, false.</returns>
		public override bool Equals(object obj)
		{
			return Normal2D.Equals(this, obj);
		}
		
		/// <summary>
		/// Returns a hash code for the current Object.
		/// </summary>
		/// <returns>A hash code for the current Object.</returns>
		public override int GetHashCode()
		{
			return this.Nx.GetHashCode() ^ this.Ny.GetHashCode();
		}

		#endregion

		// ===== Conversions ==============================================================
		#region Conversion

		// ===== Normal2D -> Normal ===============
		/// <summary>
		/// Implicit conversion Normal2D -> Normal3D
		/// </summary>
		/// <param name="n">Normal2D</param>
		/// <returns>Normal3D</returns>
		public static implicit operator Normal3D(Normal2D n) 
		{
			return new Normal3D(n.Nx, n.Ny, 0);
		} // Normal2D -> Normal3D

		/// <summary>
		/// Implicit conversion Normal2D -> NormalND
		/// </summary>
		/// <param name="n">Normal2D</param>
		/// <returns>NormalND</returns>
		public static implicit operator NormalND(Normal2D n) 
		{
			return new NormalND(new double[] {n.Nx, n.Ny});
		} // Normal2D -> NormalND

		// ===== Normal2D -> Vector ===============
		/// <summary>
		/// Implicit conversion Normal2D -> Zcu.Mve.Numerics.Vector2D
		/// </summary>
		/// <param name="n">Normal2D</param>
		/// <returns>Vector2D</returns>
		public static implicit operator Zcu.Mve.Numerics.Vector2D(Normal2D n) 
		{
			return new Zcu.Mve.Numerics.Vector2D(n.Nx, n.Ny);
		} // Normal2D -> Vector2D

		/// <summary>
		/// Implicit conversion Normal2D -> Vector3D
		/// </summary>
		/// <param name="n">Normal2D</param>
		/// <returns>Vector3D</returns>
		public static implicit operator Zcu.Mve.Numerics.Vector3D(Normal2D n) 
		{
			return new Zcu.Mve.Numerics.Vector3D(n.Nx, n.Ny, 0);
		} // Normal2D -> Vector3D

		/// <summary>
		/// Implicit conversion Normal2D -> VectorND
		/// </summary>
		/// <param name="n">Normal2D</param>
		/// <returns>VectorND</returns>
		public static implicit operator Zcu.Mve.Numerics.VectorND(Normal2D n) 
		{
			return new Zcu.Mve.Numerics.VectorND(new double[] {n.Nx, n.Ny});
		} // Normal2D -> VectorND

		// ===== Vector -> Normal2D ===============
		/// <summary>
		/// Explicit conversion Vector2D -> Normal2D.
		/// Vector is normalized.
		/// </summary>
		/// <param name="v">Vector2D</param>
		/// <returns>Normal2D</returns>
		public static explicit operator Normal2D(Zcu.Mve.Numerics.Vector2D v) 
		{
			// velikost vektoru (absolutni hodnota)
			double abs = Math.Sqrt(v.X * v.X + v.Y * v.Y);
			if (abs != 0.0)
				return new Normal2D(v.X / abs, v.Y / abs);
			else
				return new Normal2D(0.0, 0.0);
		} // Vector2D -> Normal2D

		/// <summary>
		/// Explicit conversion Vector3D -> Normal2D
		/// Z-coordinate is ignored. Vector is normalized. 
		/// </summary>
		/// <param name="v">Vector2D</param>
		/// <returns>Normal2D</returns>
		public static explicit operator Normal2D(Zcu.Mve.Numerics.Vector3D v) 
		{
			Zcu.Mve.Numerics.Vector2D v2 = new Zcu.Mve.Numerics.Vector2D(v.X, v.Y);
			return (Normal2D) v2;
		} // Vector3D -> Normal2D

		/// <summary>
		/// Explicit conversion VectorND -> Normal2D
		/// Truncates 3th, 4th ... coordinate. Vector is normalized. 
		/// </summary>
		/// <param name="v">VectorND</param>
		/// <returns>Normal2D</returns>
		public static explicit operator Normal2D(Zcu.Mve.Numerics.VectorND v) 
		{
			if (v.Values.Length == 1)
				return new Normal2D(1, 0);
			else
			{
				Zcu.Mve.Numerics.Vector2D v2 = new Zcu.Mve.Numerics.Vector2D(v.Values[0], v.Values[1]);
				return (Normal2D) v2;
			}
		} // VectorND -> Normal2D

		#endregion

		// ===== Operators ================================================================
		#region Operators

		/// <summary>
		/// The equality operator (==) returns true if the values of its operands are equal, false otherwise.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>True if the values of operands are equal, false otherwise.</returns>
		public static bool operator ==(Normal2D a, Normal2D b)
		{
			return Normal2D.Equals(a, b);
		}
		
		/// <summary>
		/// The inequality operator (!=) returns false if its operands are equal, true otherwise.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>False if the values of operands are equal, true otherwise.</returns>
		public static bool operator !=(Normal2D a, Normal2D b) 
		{
			return !(a == b);
		}

		#endregion

	} // Normal2D
} // namespace
