using System;
using System.Globalization;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Class represent 3D normal vector defined by 3 elements - Nx, Ny, Nz.
	/// </summary>
	public struct Normal3D : INormal
	{
		/// <summary>
		/// x-normal value.
		/// </summary>
		public double Nx;
		/// <summary>
		/// y-normal value.
		/// </summary>
		public double Ny;
		/// <summary>
		/// z-normal value.
		/// </summary>
		public double Nz;

		/// <summary>
		/// The toleration for equality operator.
		/// </summary>
		public static double Epsilon = Globals.EpsilonDoubleImplicit;

		/// <summary>
		/// Creates a structure and fill in given values
		/// </summary>
		/// <param name="nx">x-normal value.</param>
		/// <param name="ny">y-normal value.</param>
		/// <param name="nz">z-normal value.</param>
		public Normal3D(double nx, double ny, double nz)
		{
			this.Nx = nx;
			this.Ny = ny;
			this.Nz = nz;
		} // Normal3D()

		/// <summary>
		/// Method load data from Xml file to structures of curent data object.
		/// </summary>
		/// <param name="xmlTextReader">XML stram to read</param>
		public void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
			string[] split = xmlTextReader.ReadString().Trim().Split(' ');
			this.Nx = Double.Parse(split[0], nfi);
			this.Ny = Double.Parse(split[1], nfi);
			this.Nz = Double.Parse(split[2], nfi);
		} // ReadData()

		/// <summary>
		/// Method write data to Xml file.
		/// </summary>
		/// <param name="xmlTextWriter">Data stream to write.</param>
		public void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
			xmlTextWriter.WriteString(this.Nx.ToString(nfi) + " " 
				+ this.Ny.ToString(nfi) + " "
				+ this.Nz.ToString(nfi));
		} // WriteData()

		/// <summary>
		/// Deep copy of data object and its subelements.
		/// </summary>
		/// <returns>same data, different memmory space.</returns>
		public IDataObject DeepCopy()
		{
			return new Normal3D(this.Nx, this.Ny, this.Nz);
		} // DeepCopy()

		/// <summary>
		/// Method check whether absolute value equals 1 (+- NormalND.EPSILON). 
		/// If no, normal is normalized.
		/// </summary>
		public void Normalize()
		{
			double absValue = NormalND.AbsValue(this.Nx, this.Ny, this.Nz);
			this.Nx = this.Nx / absValue;
			this.Ny = this.Ny / absValue;
			this.Nz = this.Nz / absValue;
		} // Normalize();

		/// <summary>
		/// Check whether normal absolute value equals 1 +- NormalND.EPSILON.
		/// </summary>
		/// <returns>Returns true if normal absolute value equals 1 +- NormalND.EPSILON.</returns>
		public bool CheckConsistence()
		{
			double absValue = NormalND.AbsValue(this.Nx, this.Ny, this.Nz);
			if (absValue < 1 - NormalND.EPSILON || absValue > 1 + NormalND.EPSILON)
				return false;
			else
				return true;
		} // CheckConsistence()

		/// <summary>
		/// Get number of elements of normal.
		/// </summary>
		public int Length
		{
			get
			{
				return 3;
			}
		}

		/// <summary>
		/// Get/Set element of normal.
		/// </summary>
		public double this [int index]
		{
			get
			{
				switch(index)
				{
					case 0:
						return this.Nx;
					case 1:
						return this.Ny;
					case 2:
						return this.Nz;
					default:
						throw new IndexOutOfRangeException(this.GetType().FullName + ": Dimension exceeded.");
				}
			}
			set
			{
				switch(index)
				{
					case 0:
						this.Nx = value;
						break;
					case 1:
						this.Ny = value;
						break;
					case 2:
						this.Nz = value;
						break;
					default:
						throw new IndexOutOfRangeException(this.GetType().FullName + ": Dimension exceeded.");
				}
			}
		}

		/// <summary>
		/// Returns distance from beginning of coordinate system.
		/// </summary>
		/// <returns>Distance from beginning of coordinate system.</returns>
		public double Abs()
		{
			return Math.Sqrt(this.Nx * this.Nx + this.Ny * this.Ny + this.Nz * this.Nz);
		} // Abs()

		/// <summary>
		/// Returns array of assembly name.
		/// </summary>
		public string [] LibNames
		{
			get
			{
				return new string [] {Globals.GetLibraryName(this.GetType())};
			}
		}

		// ===== Methods of the class Object ==============================================
		#region Overrided methods of the class Object

		/// <summary>
		/// Returns a String that represents the current Object.
		/// </summary>
		/// <returns>String that represents the current Object.</returns>
		override public string ToString()
		{
			return "Normal3D: [" + this.Nx + "; " + this.Ny + "; " + this.Nz + "]";
		} // ToString()

		/// <summary>
		/// Determines whether the specified Objects are equal.
		/// </summary>
		/// <param name="a">First Object.</param>
		/// <param name="b">Second Object.</param>
		/// <returns>True if objects are equal. Otherwise, false.</returns>
		public static new bool Equals(object a, object b)
		{
			if (a.GetType() != typeof(Normal3D) || b.GetType() != typeof(Normal3D))
				return false;
		      
			return Math.Abs(((Normal3D)a).Nx - ((Normal3D)b).Nx) <= Normal3D.Epsilon
				&& Math.Abs(((Normal3D)a).Ny - ((Normal3D)b).Ny) <= Normal3D.Epsilon
				&& Math.Abs(((Normal3D)a).Nz - ((Normal3D)b).Nz) <= Normal3D.Epsilon;
		}
		
		/// <summary>
		/// Determines whether the specified Object is equal to the current Object.
		/// </summary>
		/// <param name="obj">Specified Object to compare with the current Object.</param>
		/// <returns>True if the specified Object is equal to the current Object. Otherwise, false.</returns>
		public override bool Equals(object obj)
		{
			return Normal3D.Equals(this, obj);
		}
		
		/// <summary>
		/// Returns a hash code for the current Object.
		/// </summary>
		/// <returns>A hash code for the current Object.</returns>
		public override int GetHashCode()
		{
			return this.Nx.GetHashCode() ^ this.Ny.GetHashCode() ^ this.Nz.GetHashCode();
		}

		#endregion

		// ===== Conversions ==============================================================
		#region Conversion

		// ===== Normal3D -> Normal ===============
		/// <summary>
		/// Explicit conversion Normal3D -> Normal2D
		/// </summary>
		/// <param name="n">Normal3D</param>
		/// <returns>Normal2D</returns>
		public static explicit operator Normal2D(Normal3D n) 
		{
			return new Normal2D(n.Nx, n.Ny);
		} // Normal3D -> Normal2D

		/// <summary>
		/// Implicit conversion Normal3D -> NormalND
		/// </summary>
		/// <param name="n">Normal3D</param>
		/// <returns>NormalND</returns>
		public static implicit operator NormalND(Normal3D n) 
		{
			return new NormalND(new double[] {n.Nx, n.Ny, n.Nz});
		} // Normal3D -> NormalND

		// ===== Normal3D -> Vector ===============
		/// <summary>
		/// Explicit conversion Normal3D -> Vector2D
		/// </summary>
		/// <param name="n">Normal3D</param>
		/// <returns>Vector2D</returns>
		public static explicit operator Zcu.Mve.Numerics.Vector2D(Normal3D n) 
		{
			return new Zcu.Mve.Numerics.Vector2D(n.Nx, n.Ny);
		} // Normal3D -> Vector2D

		/// <summary>
		/// Implicit conversion Normal3D -> Vector3D
		/// </summary>
		/// <param name="n">Normal3D</param>
		/// <returns>Vector3D</returns>
		public static implicit operator Zcu.Mve.Numerics.Vector3D(Normal3D n) 
		{
			return new Zcu.Mve.Numerics.Vector3D(n.Nx, n.Ny, n.Nz);
		} // Normal3D -> Vector3D

		/// <summary>
		/// Implicit conversion Normal3D -> VectorND
		/// </summary>
		/// <param name="n">Normal3D</param>
		/// <returns>VectorND</returns>
		public static implicit operator Zcu.Mve.Numerics.VectorND(Normal3D n) 
		{
			return new Zcu.Mve.Numerics.VectorND(new double[] {n.Nx, n.Ny, n.Nz});
		} // Normal3D -> VectorND

		// ===== Vector -> Normal3D ===============
		/// <summary>
		/// Explicit conversion Vector2D -> Normal3D
		/// </summary>
		/// <param name="v">Vector2D</param>
		/// <returns>Normal3D</returns>
		public static explicit operator Normal3D(Zcu.Mve.Numerics.Vector2D v) 
		{
			Normal2D tmp = (Normal2D) v;
			return new Normal3D(tmp.Nx, tmp.Ny, 0);
		} // Vector2D -> Normal3D

		/// <summary>
		/// Explicit conversion Vector3D -> Normal3D
		/// </summary>
		/// <param name="v">Vector3D</param>
		/// <returns>Normal3D</returns>
		public static explicit operator Normal3D(Zcu.Mve.Numerics.Vector3D v) 
		{
			// velikost vektoru (absolutni hodnota)
			double abs = Math.Sqrt(Math.Pow(v.X, 2) + Math.Pow(v.Y, 2) + Math.Pow(v.Z, 2));
			return new Normal3D(v.X / abs, v.Y / abs, v.Z / abs);
		} // Vector3D -> Normal3D

		/// <summary>
		/// Explicit conversion VectorND -> Normal3D
		/// </summary>
		/// <param name="v">VectorND</param>
		/// <returns>Normal3D</returns>
		public static explicit operator Normal3D(Zcu.Mve.Numerics.VectorND v) 
		{
			if (v.Values.Length == 1)
				return new Normal3D(1, 0, 0);
			else if (v.Values.Length == 2)
			{
				Zcu.Mve.Numerics.Vector2D v2 = new Zcu.Mve.Numerics.Vector2D(v.Values[0], v.Values[1]);
				return (Normal3D) v2;
			}
			else
			{
				Zcu.Mve.Numerics.Vector3D v3 = new Zcu.Mve.Numerics.Vector3D(v.Values[0], v.Values[1], v.Values[2]);
				return (Normal3D) v3;
			}
		} // VectorND -> Normal3D

		#endregion

		// ===== Operators ================================================================
		#region Operators

		/// <summary>
		/// The equality operator (==) returns true if the values of its operands are equal, false otherwise.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>True if the values of operands are equal, false otherwise.</returns>
		public static bool operator ==(Normal3D a, Normal3D b)
		{
			return Normal3D.Equals(a, b);
		}
		
		/// <summary>
		/// The inequality operator (!=) returns false if its operands are equal, true otherwise.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>False if the values of operands are equal, true otherwise.</returns>
		public static bool operator !=(Normal3D a, Normal3D b) 
		{
			return !(a == b);
		}

		/// <summary>
		/// Addition of two normals.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>Resulting sum (a + b).</returns>
		public static Normal3D operator +(Normal3D a, Normal3D b)
		{
			return(new Normal3D(a.Nx+b.Nx,a.Ny+b.Ny,a.Nz+b.Nz));
		}

		/// <summary>
		/// Subtraction of two normals.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>Resulting difference (a - b).</returns>
		public static Normal3D operator -(Normal3D a, Normal3D b)
		{
			return(new Normal3D(a.Nx-b.Nx, a.Ny-b.Ny, a.Nz-b.Nz));
		}

		/// <summary>
		/// Multiplication by constant number.
		/// </summary>
		/// <param name="p">Normal to be multiplied.</param>
		/// <param name="m">Multiplicator.</param>
		/// <returns>Resulting normal (m*p.Nx, m*p.y, m*p.z)</returns>
		public static Normal3D operator *(Normal3D p, double m)
		{
			return(new Normal3D(p.Nx*m, p.Ny*m, p.Nz*m));
		}

		/// <summary>
		/// Multiplication by a scalar.
		/// </summary>
		/// <param name="p">Point to be multiplied.</param>
		/// <param name="m">Multiplicator.</param>
		/// <returns>Resulting point (m*p.Nx, m*p.y, m*p.z).</returns>
		public static Normal3D operator *(double m, Normal3D p)
		{
			return(new Normal3D(p.Nx*m, p.Ny*m, p.Nz*m));
		}

		/// <summary>
		/// Division by a scalar.
		/// </summary>
		/// <param name="p">Point to be divided.</param>
		/// <param name="m">Divider.</param>
		/// <returns>Resulting point (p.Nx/m, p.y/m, p.z/m).</returns>
		public static Normal3D operator /(Normal3D p, double m)
		{
			return(new Normal3D(p.Nx/m, p.Ny/m, p.Nz/m));
		}

		#endregion

		// ===== Operations ===============================================================
		#region Operations

		/// <summary>
		/// Returns euclidean distance between two normals.
		/// </summary>
		/// <param name="p">Argument normal to which distance will be computed.</param>
		/// <returns>Euclidean distance between the normal.</returns>
		public double DistanceTo(Normal3D p)
		{
			return(Math.Sqrt((p.Nx-this.Nx)*(p.Nx-this.Nx)+(p.Ny-this.Ny)*(p.Ny-this.Ny)+(p.Nz-this.Nz)*(p.Nz-this.Nz)));
		}

		/// <summary>
		/// Returns squared euclidean distance between two normals
		/// (is slightly faster than the DistanceTo function as no square root is computed).
		/// </summary>
		/// <param name="p">Argument normal to which distance will be computed.</param>
		/// <returns>Square euclidean distance between the normals.</returns>
		public double SquareDistanceTo(Normal3D p)
		{
			return((p.Nx-this.Nx)*(p.Nx-this.Nx)+(p.Ny-this.Ny)*(p.Ny-this.Ny)+(p.Nz-this.Nz)*(p.Nz-this.Nz));
		}

		/// <summary>
		/// Returns the dot product of two normals.
		/// </summary>
		/// <param name="arg">Second argument of the dot product operation.</param>
		/// <returns>Scalar value of this.arg.</returns>
		public double DotProduct(Normal3D arg)
		{
			return(this.Nx*arg.Nx+this.Ny*arg.Ny+this.Nz*arg.Nz);
		}

		/// <summary>
		/// Returns the cross product of two normals.
		/// </summary>
		/// <param name="arg">Second argument of the cross product operation.</param>
		/// <returns>Normal3D representing the resulting vector.</returns>
		public Normal3D CrossProduct(Normal3D arg)
		{
			return(new Normal3D(this.Ny*arg.Nz-this.Nz*arg.Ny,this.Nz*arg.Nx-this.Nx*arg.Nz,this.Nx*arg.Ny-this.Ny*arg.Nx));
		}

		#endregion

	} // class Normal3D
} // namespace


