using System;
using System.Collections;
using System.Xml;
using System.Diagnostics;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Class represents a Slice. This is typical composed cell. 
	/// It contains number of PolyLine elements.
	/// </summary>
	public class Slice : DataObject, ICell
	{
		/// <summary>
		/// ArrayList of polyLines.
		/// </summary>
		private ArrayList polyLineList;

		/// <summary>
		/// Implicit constructor - creates an empty slice.
		/// </summary>
		public Slice()
		{
			this.polyLineList = new ArrayList();
		}  // Slice()

		/// <summary>
		/// Get the list of polyLines.
		/// </summary>
		public ICollection PolyLineList
		{
			get
			{
				return this.polyLineList;
			}
		} // PolyLineList

		/// <summary>
		/// Get the count of polyLines.
		/// </summary>
		public int PolyLineCount
		{
			get
			{
				return this.polyLineList.Count;
			}
		} // PolyLineCount

		/// <summary>
		/// Add the PolyLine and return its index in the Slice.
		/// </summary>
		/// <param name="polyLine">PolyLine.</param>
		/// <returns>Index of the just added PolyLine.</returns>
		public int AddPolyLine(PolyLine polyLine)
		{
			return this.polyLineList.Add(polyLine);
		} // AddPolyLine()

		/// <summary>
		/// Get PolyLine.
		/// </summary>
		/// <param name="index">Index of PolyLine.</param>
		/// <returns>PolyLine with given index</returns>
		public PolyLine GetPolyLine(int index)
		{
			if (index < this.polyLineList.Count && index >= 0)
				return (PolyLine) this.polyLineList[index];
			else
				throw new MveVisualizationException(this.GetType().Name + ": index " + index + " is out of range <0, " + (this.polyLineList.Count - 1) + ">");
		} // GetPolyLine()

		/// <summary>
		/// Set PolyLine.
		/// </summary>
		/// <param name="index">Index of PolyLine.</param>
		/// <param name="val">PolyLine with given index.</param>
		public void SetPolyLine(int index, PolyLine val)
		{
			if (index < this.polyLineList.Count && index >= 0)
				this.polyLineList[index] = val;
			else
				throw new MveVisualizationException(this.GetType().Name + ": index " + index + " is out of range <0, " + (this.polyLineList.Count - 1) + ">");
		} // SetPolyLine()

		/*
				/// <summary>
				/// Indexer pro pstup k jednotlivm PolyLines ezu pomoc index.
				/// Prvn PolyLine m index 0.
				/// </summary>
				public PolyLine this[int index]
				{
					get
					{
						if (index < this.polyLineList.Count && index >= 0)
							return (PolyLine) this.polyLineList[index];
						else
							throw new MveVisualizationException(this.GetType().Name + ": index " + index + " is out of range <0, " + (this.polyLineList.Count - 1) + ">");
					}
					set
					{
						if (index < this.polyLineList.Count && index >= 0)
							this.polyLineList[index] = value;
						else
							throw new MveVisualizationException(this.GetType().Name + ": index " + index + " is out of range <0, " + (this.polyLineList.Count - 1) + ">");
					}
				} // this[int]
		*/

		/// <summary>
		/// Read data from Xml file to current structure.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		public override void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			// poet vech PolyLines tvocch ez
			int pcount = Int32.Parse(xmlTextReader.GetAttribute("pcount").Trim());
			// poet aktuln natench PolyLines
			int i = 0;
			// natej dokud nejsou nateny vechny PolyLines nebo nen dosaeno konce xml souboru
			while(xmlTextReader.Read() && i < pcount)
			{
				if (xmlTextReader.NodeType == XmlNodeType.Element && xmlTextReader.Name == "p")
				{
					PolyLine polyLine = new PolyLine();
					polyLine.ReadData(xmlTextReader);
					this.AddPolyLine(polyLine);
					i++;
				}
			} // while
		} // ReadData()

		/// <summary>
		/// Write current data to Xml file.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		public override void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			// poet vech PolyLines tvocch ez je zapsn jako atribut aktulnho uzlu (rodie samotnch dat)
			xmlTextWriter.WriteAttributeString("pcount", this.polyLineList.Count.ToString());
			foreach(PolyLine polyLine in this.polyLineList)
			{	// potomky aktulnho uzlu jsou uzly 'p' obsahujc body jednotlivch PolyLines
				xmlTextWriter.WriteStartElement("p");
				polyLine.WriteData(xmlTextWriter);
				xmlTextWriter.WriteEndElement();
			}
		} // WriteData()

		/// <summary>
		/// Deep copy of data object and its subelements.
		/// </summary>
		/// <returns>Same data, different memory space.</returns>
		public override IDataObject DeepCopy()
		{
			Slice tmpSlice = new Slice();
			foreach(PolyLine polyLine in this.polyLineList)
				tmpSlice.AddPolyLine((PolyLine)polyLine.DeepCopy());
			return tmpSlice;
		} // DeepCopy()

		/// <summary>
		/// Check the consistency of current data structure.
		/// </summary>
		/// <returns>True if structure is consistent.</returns>
		public override bool CheckConsistence()
		{
			bool isOK = true;
			for (int i = 0; i < this.polyLineList.Count; i++)
				if (GetPolyLine(i).CheckConsistence() == false)
					isOK = false;
			return isOK;
		} // CheckConsistence()

		/// <summary>
		/// Returns a String that represents the current Object.
		/// </summary>
		/// <returns>String that represents the current Object.</returns>
		public override string ToString()
		{
			string pom = "Slice (" + this.polyLineList.Count + " polylines):";
			if (this.polyLineList.Count == 0)
				return pom += " {}";
			foreach (PolyLine polyLine in this.polyLineList)
				pom += "\n\t" + polyLine.ToString();
			return pom;
		} // ToString()

		/// <summary>
		/// Composed cell. Do not have mening in case of Slice.
		/// </summary>
		public int this[int index]
		{
			get
			{
				return 0;
			}
			set
			{
			}
		}

		/// <summary>
		/// Composed cell. Do not have mening in case of Slice.
		/// </summary>
		public int PointCount
		{
			get
			{
				return 0;
			}
		} // PointCount

	} // class Slice
} // namespace
