using System;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// General tranformation of a 3D object.
	/// </summary>
	public class Transform3D : Visualization.AbstractTransform
	{
		/// <summary>
		/// Float matrix 4 x 4 representing 3D-transformation.
		/// </summary>
		private float [,] matrix = new float [4, 4];

		/// <summary>
		/// The toleration for equality operator.
		/// </summary>
		public static double Epsilon = Globals.EpsilonFloatImplicit;

		/// <summary>
		/// Creates a Transform3D object and initialize it (set identity).
		/// </summary>
		public Transform3D()
		{
			SetIdentity();
		}

		/// <summary>
		/// Set identity transformation.
		/// </summary>
		public void SetIdentity()
		{
			for (int i = 0; i < 4; i++)
			{
				for (int j = 0; j < 4; j++)
				{
					matrix[i, j] = 0.0f;
				}
			}
			for (int i = 0; i < 4; i++)
			{
				matrix[i, i] = 1.0f;
			}
		}
    
		/// <summary>
		/// Add scaling to current transformation.
		/// TODO: Optimize it!
		/// </summary>
		/// <param name="dx">Scaling in x-axis. 1 means no scale.</param>
		/// <param name="dy">Scaling in y-axis. 1 means no scale.</param>
		/// <param name="dz">Scaling in z-axis. 1 means no scale.</param>
		public void AddScale(float dx, float dy, float dz)
		{
			Transform3D t = new Transform3D();
			t.SetScale(dx, dy, dz);
			this.matrix = (this * t).matrix;
		}

		/// <summary>
		/// Set curent matrix to scaling matrix of given values.
		/// </summary>
		/// <param name="dx">Scaling in x-axis. 1 means no scale.</param>
		/// <param name="dy">Scaling in y-axis. 1 means no scale.</param>
		/// <param name="dz">Scaling in z-axis. 1 means no scale.</param>
		public void SetScale(float dx, float dy, float dz)
		{
			SetIdentity();
			matrix[0, 0] = dx;
			matrix[1, 1] = dy;
			matrix[2, 2] = dz;
		}

		/// <summary>
		/// Noncumulative rotation about axis X measured in radians.
		/// </summary>
		/// <param name="angle">Rotation angle masured in radians.</param>
		public void SetRotaionX(float angle)
		{
			SetIdentity();
			matrix[1, 1] = matrix[2, 2] = (float) Math.Cos(angle);
			matrix[2, 1] = (float) Math.Sin(angle);      
			matrix[1, 2] = - matrix[2, 1];
		}
    
		/// <summary>
		/// Cumulative rotation about axis X.
		/// </summary>
		/// <param name="angle">Rotation angle masured in radians.</param>
		public void AddRotationX(float angle)
		{
			Transform3D rot = new Transform3D();
			rot.SetRotaionX(angle);
			this.matrix = (this * rot).matrix;
		}

		/// <summary>
		/// Noncumulative rotation about axis Y measured in radians.
		/// </summary>
		/// <param name="angle">Rotation angle masured in radians.</param>
		public void SetRotaionY(float angle)
		{
			SetIdentity();
			matrix[0, 0] = matrix[2, 2] = (float) Math.Cos(angle);
			matrix[2, 0] = (float) Math.Sin(angle);      
			matrix[0, 2] = - matrix[2, 0];
		}
    
		/// <summary>
		/// Cumulative rotation about axis Y.
		/// </summary>
		/// <param name="angle">Rotation angle masured in radians.</param>
		public void AddRotationY(float angle)
		{
			Transform3D rot = new Transform3D();
			rot.SetRotaionY(angle);
			this.matrix = (this * rot).matrix;
		}
    
		/// <summary>
		/// Noncumulative rotation about axis Z measured in radians.
		/// </summary>
		/// <param name="angle">Rotation angle masured in radians.</param>
		public void SetRotaionZ(float angle)
		{
			SetIdentity();
			matrix[0, 0] = matrix[1, 1] = (float) Math.Cos(angle);
			matrix[1, 0] = (float) Math.Sin(angle);      
			matrix[0, 1] = - matrix[1, 0];
		}

		/// <summary>
		/// Cumulative rotation about axis Z.
		/// </summary>
		/// <param name="angle">Rotation angle masured in radians.</param>
		public void AddRotationZ(float angle)
		{
			Transform3D rot = new Transform3D();
			rot.SetRotaionZ(angle);
			this.matrix = (this * rot).matrix;
		}

		/// <summary>
		/// Noncumlative translation.
		/// </summary>
		/// <param name="x">Translation in x-axis.</param>
		/// <param name="y">Translation in y-axis.</param>
		/// <param name="z">Translation in z-axis.</param>
		public void SetTranslation(float x, float y, float z)
		{
			Transform3D t = new Transform3D();
			matrix[0, 3] = x;
			matrix[1, 3] = y;
			matrix[2, 3] = z;
		}

		/// <summary>
		/// Cumulative translation
		/// </summary>
		/// <param name="x">Translation in x-axis.</param>
		/// <param name="y">Translation in y-axis.</param>
		/// <param name="z">Translation in z-axis.</param>
		public void AddTranslation(float x, float y, float z)
		{
			Transform3D t = new Transform3D();
			t.SetTranslation(x, y, z);
			this.matrix = (this * t).matrix;
		}    

		/// <summary>
		/// Transform the point.
		/// </summary>
		/// <param name="op">Point to be transform. This parameter won't be changed.</param>
		/// <returns>New instance of transformated point.</returns>
		public Point3D Transform(Point3D op)
		{
			Point3D tp = new Point3D();
      
			tp.X = matrix[0, 0] * op.X + matrix[0, 1] * op.Y + matrix[0, 2] * op.Z + matrix[0, 3];
			tp.Y = matrix[1, 0] * op.X + matrix[1, 1] * op.Y + matrix[1, 2] * op.Z + matrix[1, 3];
			tp.Z = matrix[2, 0] * op.X + matrix[2, 1] * op.Y + matrix[2, 2] * op.Z + matrix[2, 3];
			// homogeni slozka ignorovana

			return tp;
		}

		/// <summary>
		/// Composition of transformation. Matrix multiplication.
		/// </summary>
		/// <param name="t1"></param>
		/// <param name="t2"></param>
		/// <returns></returns>
		public static Transform3D operator * (Transform3D t1, Transform3D t2)
		{
			Transform3D t = new Transform3D();
			float sum = 0.0f;
      
			for (int i = 0; i < 4; i++)
			{
				for (int j = 0; j < 4; j++)
				{
					sum = 0.0f;
					for (int k = 0; k < 4; k++)
					{
						sum += t1[i, k] * t2[k, j];
					}
					t[i, j] = sum;
				}
			}
			return t;
		}

		/// <summary>
		/// Full access to particular matrix elements.
		/// </summary>
		public float this [int i, int j]
		{
			set
			{
				matrix[i, j] = value;
			}
			get
			{
				return matrix[i, j];
			}
		}
	  
		/// <summary>
		/// Check the consistency of current data structure.
		/// </summary>
		/// <returns>True if structure is consistent.</returns>
		public override bool CheckConsistence()
		{
			// TODO: mozna nejaky determinant-test a nasledny warning
			return true;
		}
  
		/// <summary>
		/// Deep copy of data object and its subelements.
		/// </summary>
		/// <returns>Same data, different memory space.</returns>
		public override Zcu.Mve.Core.IDataObject DeepCopy()
		{
			Transform3D t = new Transform3D();

			for (int i = 0; i < 4; i++)
			{
				for (int j = 0; j < 4; j++)
				{
					t.matrix[i, j] = this.matrix[i, j];
				}
			}   
			return null;
		}
  
		/// <summary>
		/// Read data from Xml file to current structure.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		public override void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			string [] lines = xmlTextReader.ReadString().Trim().Split(';');
      
			if (lines.Length != 4)
				throw new MveException("Incorrect number of lines in Transform3D.");

			for (int i = 0; i < 4; i++)
			{
				string [] elems = lines[i].Trim().Split(' ');
				if (elems.Length != 4)
					throw new MveException("Incorrect number of elements in line " + i + " in Transform3D.");
        
				for (int j = 0; j < 4; j++)
				{
					this.matrix[i, j] = Int32.Parse(elems[j].Trim(), Globals.nfi);
				}
			}
		}

		/// <summary>
		/// Write current data to Xml file.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		public override void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			string str = "";
			for (int i = 0; i < 4; i++)
			{
				for (int j = 0; j < 4; j++)
				{
					str += this.matrix[i, j].ToString(Globals.nfi);
					if (j < 3)
						str += " ";
				}
				if (i < 3)
					str += "; ";
			}
			xmlTextWriter.WriteString(str);
		}

		// ===== Methods of the class Object ==============================================
		#region Overrided methods of the class Object

		/// <summary>
		/// Returns a String that represents the current Object.
		/// </summary>
		/// <returns>String that represents the current Object.</returns>
		public override string ToString()
		{
			string str = "Transform3D:\n";
			for (int i = 0; i < 4; i++)
				str += "   [" + matrix[i, 0] + ", " + matrix[i, 1] + ", " + matrix[i, 2] + ", " + matrix[i, 3] + "]\r\n";
			return str.Substring(0, str.Length - 2);
		}

		/// <summary>
		/// Determines whether the specified Objects are equal.
		/// </summary>
		/// <param name="a">First Object.</param>
		/// <param name="b">Second Object.</param>
		/// <returns>True if objects are equal. Otherwise, false.</returns>
		public static new bool Equals(object a, object b)
		{
			if (a.GetType() != typeof(Transform3D) || b.GetType() != typeof(Transform3D))
				return false;

			Transform3D x = (Transform3D) a;
			Transform3D y = (Transform3D) b;

			for (int i = 0; i < 4; i++)
				for (int j = 0; j < 4; j++)
					if ((float)Math.Abs(x[i,j] - y[i,j]) > Transform3D.Epsilon)
						return false;

			return true;
		}
		
		/// <summary>
		/// Determines whether the specified Object is equal to the current Object.
		/// </summary>
		/// <param name="obj">Specified Object to compare with the current Object.</param>
		/// <returns>True if the specified Object is equal to the current Object. Otherwise, false.</returns>
		public override bool Equals(object obj)
		{
			return Transform3D.Equals(this, obj);
		}
		
		/// <summary>
		/// Returns a hash code for the current Object.
		/// </summary>
		/// <returns>A hash code for the current Object.</returns>
		public override int GetHashCode()
		{
			int hashCode = this[0,0].GetHashCode() 
				^ this[0,1].GetHashCode()
				^ this[0,2].GetHashCode()
				^ this[0,3].GetHashCode();

			for (int i = 1; i < 3; i++)
				for (int j = 0; j < 3; j++)
					hashCode = hashCode ^ this[i,j].GetHashCode();

			return hashCode;
		}

		#endregion

		// ===== Operators ================================================================
		#region Operators

		/// <summary>
		/// The equality operator (==) returns true if the values of its operands are equal, false otherwise.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>True if the values of operands are equal, false otherwise.</returns>
		public static bool operator ==(Transform3D a, Transform3D b)
		{
			return Transform3D.Equals(a, b);
		}
		
		/// <summary>
		/// The inequality operator (!=) returns false if its operands are equal, true otherwise.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>False if the values of operands are equal, true otherwise.</returns>
		public static bool operator !=(Transform3D a, Transform3D b) 
		{
			return !(a == b);
		}

		#endregion

	} // class Transform3D
} // namespace