using System;
using System.Globalization;
using System.Collections;
using System.IO;
using System.Diagnostics;
using System.Text.RegularExpressions;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Module for loading of triangle mesh from .tri format file.
	/// Support [Vertices], [Points], [Vertices' Normals], [Triangles' Normals]
	/// </summary>
	[ModuleInfo("Miroslav Vavruka", 
		 "Module for loading of triangle mesh from .tri format file.", 
		 Assemblied = "2004-07-01",
		 IconName = "TriangleLoader.ico")]
	[PortInfo("Output", "Loaded Mesh.")]
	public class TriangleLoader : Zcu.Mve.Core.Module
	{
		/// <summary>
		/// Name of the file containing data.
		/// </summary>
		private string fileName;
        		
		/// <summary>
		/// Creates module and output port, initialize fileName.
		/// </summary>
		public TriangleLoader()
		{
			AddOutPort("Output", typeof (Visualization.TriangleMesh));
			this.fileName = "";
		} // TriangleLoader()

		/// <summary>
		/// Get/Set the name of the file containing data.
		/// </summary>
		public string FileName
		{
			get
			{
				return this.fileName;
			}
			set
			{
				this.fileName = value;
			}
		} // FileName

		/// <summary>
		/// Read the configuration data from XML file.
		/// </summary>
		/// <param name="config">XmlElement containing the configuration data.</param>
		public override void ReadConfig(System.Xml.XmlElement config)
		{
			this.FileName = config.InnerText.Trim();
		} // ReadConfig()

		/// <summary>
		/// Write the configuration data in XML file.
		/// </summary>
		/// <param name="config">XmlElement containing the configuration data.</param>
		public override void WriteConfig(System.Xml.XmlElement config)
		{
			config.InnerText = this.FileName;
		} // WriteConfig()

		/// <summary>
		/// Executes TriangleLoader functionality.
		/// </summary>
		public override void Execute()
		{
			if (this.fileName == "")		// nezadano zadne jmeno souboru
				throw new MveException(this.GetType().Name + " \"" + this.Name + "\" error: Pleas enter the name of file with data.");

			StreamReader sr = null;

			try
			{
				sr = new StreamReader(fileName);

				// initializa arrraylists
				ArrayList points = new ArrayList();
				ArrayList cells = new ArrayList();
				ArrayList pointNormals = new ArrayList();
				ArrayList cellNormals = new ArrayList();

				// read data from file
				ReadTriFile(sr, points, cells, pointNormals, cellNormals);

				// create mesh and fill it with loaded data
				TriangleMesh mesh = new TriangleMesh();        
				mesh.Points = UniformDataArray.FromArrayList(typeof(Point3D), points);
				mesh.AddCells(typeof(Triangle).Name, UniformDataArray.FromArrayList(typeof(Triangle), cells));

				if (cellNormals.Count != 0)
					mesh.AddCellAttrs(typeof(Triangle).Name, typeof(Normal3D).Name, UniformDataArray.FromArrayList(typeof(Normal3D), cellNormals));

				if (pointNormals.Count != 0)
					mesh.AddPointAttrs(typeof(Normal3D).Name, UniformDataArray.FromArrayList(typeof(Normal3D), pointNormals));

				SetOutput("Output", mesh);
			} // try
			catch (FileNotFoundException ex)
			{
				Debug.WriteLine(this.GetType().Name + " '" + this.Name + "' error:");
				Debug.WriteLine(ex);
				throw new MveException(this.GetType().Name + " \"" + this.Name + "\" error: can not find file: \"" + this.fileName + "\"");
			}
			catch (DirectoryNotFoundException ex)
			{
				Debug.WriteLine(this.GetType().Name + " '" + this.Name + "' error:");
				Debug.WriteLine(ex);
				throw new MveException(this.GetType().Name + " \"" + this.Name + "\" error: can not find directory: \"" + this.fileName + "\"");
			}
			finally
			{
				if (sr != null)
					sr.Close();
			}
		} // Execute()

		/// <summary>
		/// Support method, reads data from the opened stream
		/// </summary>
		/// <param name="sr">Strem to read.</param>
		/// <param name="pts">List of points.</param>
		/// <param name="trs">List of triangles.</param>
		/// <param name="ptsNorm">List of points normals.</param>
		/// <param name="trsNorm">List of triangles normals.</param>
		private void ReadTriFile(StreamReader sr, ArrayList pts, ArrayList trs, ArrayList ptsNorm, ArrayList trsNorm)
		{
			int nLine = 1;					// cislo prave nacitane radky
			NumberFormatInfo nfi = Globals.nfi;
			string line;
			line = sr.ReadLine();

			try 
			{
				while (line != null)
				{
					switch (line.Trim())
					{
						case "[Vertices]":    // nacteni bodu
							line = sr.ReadLine();
							nLine++;
							while (line != null && (line.IndexOf("[") == -1))
							{              
								AddPoint(pts, line);
								line = sr.ReadLine();
								if (line != null) nLine++;
							}
							break;
	
						case "[Triangles]":    // nacteni trojuhelniku
							line = sr.ReadLine();
							nLine++;
							while (line != null && (line.IndexOf("[") == -1))
							{
								AddTriangle(trs, line);
								line = sr.ReadLine();
								if (line != null) nLine++;
							}
							break;

						case "[Vertices' Normals]":	 // nacteni normal bodu
							line = sr.ReadLine();
							nLine++;
							while (line != null && (line.IndexOf("[") == -1))
							{
								AddNormal(ptsNorm, line);
								line = sr.ReadLine();
								if (line != null) nLine++;
							}
							break;
						
						case "[Triangles' Normals]":	// nacteni normal trojuhelniku
							line = sr.ReadLine();
							nLine++;
							while (line != null && (line.IndexOf("[") == -1))
							{
								AddNormal(trsNorm, line);
								line = sr.ReadLine();
								if (line != null) nLine++;
							}
							break;

						default:						// nacteni radky s 'nepotrebnymi' udaji
							line = sr.ReadLine();
							nLine++;
							break;
					} // switch
				} // while
			}
			catch (Exception ex)
			{
				Debug.WriteLine(this.GetType().Name + " '" + this.Name + "' error:");
				Debug.WriteLine(ex);
				throw new MveException(this.GetType().Name + " \"" + this.Name + "\" error: Invalid data format at line: " + nLine + " in file \"" + this.fileName + "\"");
			} // catch
		}

		/// <summary>
		/// Split the lino into 3 double values (point coordinates).
		/// </summary>
		/// <param name="line">Current line loaded from the file.</param>
		/// <param name="x">x-coordinate of Point.</param>
		/// <param name="y">y-coordinate of Point.</param>
		/// <param name="z">z-coordinate of Point.</param>
		private void SplitAndCorrectLineDouble(string line, out double x, out double y, out double z)
		{
			x = y = z = 0.0;

			string [] elems = line.Trim().Split(' ', '\t');
			if (elems.Length != 3)
			{
				throw new MveException("Invalind number of elements ...");
			}
			x = Double.Parse(elems[0], Globals.nfi);
			y = Double.Parse(elems[1], Globals.nfi);
			z = Double.Parse(elems[2], Globals.nfi);

		} // SplitAndCorrectLine()

		/// <summary>
		/// Split the lino into 3 integer values (for example indices).
		/// </summary>
		/// <param name="line">Current line loaded from the file.</param>
		/// <param name="x">First value.</param>
		/// <param name="y">Second value.</param>
		/// <param name="z">Third value.</param>
		private void SplitAndCorrectLineInt32(string line, out int x, out int y, out int z)
		{
			x = y = z = 0;

			string [] elems = line.Trim().Split(' ', '\t');
			if (elems.Length != 3)
			{
				throw new MveException("Invalind number of elements ...");
			}
			x = Int32.Parse(elems[0], Globals.nfi);
			y = Int32.Parse(elems[1], Globals.nfi);
			z = Int32.Parse(elems[2], Globals.nfi);

		} // SplitAndCorrectLine()

		/// <summary>
		/// Parse line containing point coordinates and add the point into list.
		/// </summary>
		/// <param name="triangles">List of points.</param>
		/// <param name="line">Current line loaded from the file.</param>
		private void AddPoint(ArrayList triangles, string line)
		{
			double x, y, z;
			if (line == "") return;
			this.SplitAndCorrectLineDouble(line, out x, out y, out z);
			triangles.Add(new Point3D(x, y, z));
		} // InsertPoint()

		/// <summary>
		/// Parse line containing triangle indices and add the triangle into list.
		/// </summary>
		/// <param name="triangles">List of triangles.</param>
		/// <param name="line">Current line loaded from the file.</param>
		private void AddTriangle(ArrayList triangles, string line)
		{
			if (line == "") 
				return;
			int i, j, k;

			this.SplitAndCorrectLineInt32(line, out i, out j, out k);

			triangles.Add(new Triangle(i, j, k));
		} // InsertCell()

		/// <summary>
		/// Parse line containing normal values and add the normal into list.
		/// </summary>
		/// <param name="normals">List of normals.</param>
		/// <param name="line">Current line loaded from the file.</param>
		private void AddNormal(ArrayList normals, string line)
		{
			double x, y, z;
			if (line == "") return;
			this.SplitAndCorrectLineDouble(line, out x, out y, out z);
			normals.Add(new Normal3D(x, y, z));
		} // AddNormal()

		/// <summary>
		/// Invoke module setup dialog.
		/// </summary>
		/// <returns>Module setup dialog.</returns>
		public override ModuleSetup InvokeSetup()
		{
			return new TriangleLoaderSetup(this);
		} // InvokeSetup()

	} // class TriangleLoader
} // namespace
