using System;
using System.Xml;
using System.Diagnostics;
using System.Collections;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Base class of the most general data type in Visualization library. <br></br>
	/// It consists of explicit arrays of points, cells, attributes.<br></br>
	/// Points are stored in one UniformDataArray. Thence it follows only one type of points is allowed. <br></br>
	/// Cells are stored in table of UniformDataArray. Thus there is possible to store arbitrary number of cells of different types. 
	/// Each cell array is identified by a string ID. (Have to be unique in frame of one instance of UnstrGrid.)<br></br>
	/// Attribute arrays (uniform) are associated with points and cells array. 
	/// Each cell or point array can have arbitrary number of attribute arrays. 
	/// Attribute arrays are identified by a string ID. (Similar to cell arrays)
	/// </summary>
	public class UnstrGrid : DataSet
	{
		/// <summary>
		/// Array of points.
		/// </summary>
		private UniformDataArray points;
    
		/// <summary>
		/// Table of cell arrays (UniformDataArray).
		/// </summary>
		private Hashtable cells = new Hashtable();

		/// <summary>
		/// Table of point attributes (UniformDataDrray).
		/// </summary>
		private Hashtable pointAttrs = new Hashtable();

		/// <summary>
		/// Table of tables of cell attributes (UniformDataDrray).
		/// </summary>
		private Hashtable cellAttrs = new Hashtable();

		/// <summary>
		/// Maintain min-max box.
		/// </summary>
		private BoundingBox boundingBox = null;

		/// <summary>
		/// Transform of curent object.
		/// </summary>
		private AbstractTransform transform;

		/// <summary>
		/// Creates empty (no point array, no cell array and no attributes) unstructured grid.
		/// </summary>
		public UnstrGrid()
		{
			this.transform = new Transform3D();
			// this.transform.SetIdentity();
		}

		/// <summary>
		/// Returns array of cell arrays of given type.
		/// </summary>
		/// <param name="typeOfCells">Type of cells.</param>
		/// <returns>Array of cell arrays of given type.</returns>
		public UniformDataArray [] GetCells(Type typeOfCells)
		{
			ArrayList list = new ArrayList();
      
			foreach (string cellId in GetCellIds())
			{
				UniformDataArray cellArray = GetCells(cellId);
				if (cellArray.ElemType == typeOfCells)
				{
					list.Add(cellArray);
				}
			}
			return (UniformDataArray []) list.ToArray(typeof(UniformDataArray));
		}

		/// <summary>
		/// Remove array of cell attributes. If do not exists throws an MveException.
		/// </summary>
		/// <param name="cellId">Id of cell array.</param>
		/// <param name="attrId">Id of attribute array to be removed.</param>
		public void RemoveCellAttrs(string cellId, string attrId)
		{
			if (cellAttrs.ContainsKey(cellId))
			{
				Hashtable attrs = (Hashtable) cellAttrs[cellId];

				if (attrs.ContainsKey(attrId))
				{
					attrs.Remove(attrId);
				}
				else
				{
					throw new MveException("Try to remove nonexisting array of cell attributes.");
				}
			}
			else
			{
				throw new MveException("Try to remove attribute array from nonexisting array of cells.");
			}
		}

		/// <summary>
		/// Remove array of point attributes. If do not exists throws an MveException.
		/// </summary>
		/// <param name="attrId">id of array to be removed</param>
		public void RemovePointAttrs(string attrId)
		{
			if (pointAttrs.ContainsKey(attrId))
				pointAttrs.Remove(attrId);
			else
				throw new MveException("UnstrGrid: Try to remove nonexisting array of cell attributes.");
		}

		/// <summary>
		/// Add array of cell attributes.
		/// The cell array have to exists.
		/// The length have to be equal length of cells array.
		/// </summary>
		/// <param name="cellId">Id of cell array.</param>
		/// <param name="attrId">Id of attribute array to be removed.</param>
		/// <param name="attrs">Attributes to be added.</param>
		public virtual void AddCellAttrs(string cellId, string attrId, UniformDataArray attrs)
		{
			if (!cellAttrs.ContainsKey(cellId))
				throw new MveException("UnstrGrid do not contain cells with " + cellId + ".");

			if (cellAttrs.ContainsKey(attrId))
				throw new MveException("UnstrGrid: Cells " + cellId + " already contains attrid " + attrId + ".");

			if (GetCells(cellId).Count != attrs.Count)
				throw new MveException("Count of cell attributes do not agree with count of cells.");

			Hashtable table = (Hashtable) cellAttrs[cellId];
			table.Add(attrId, attrs);
		}

		/// <summary>
		/// Add uniform cell array.
		/// </summary>
		/// <param name="cellId">Unique (in curent UnstrGrid) ID of cell array.</param>
		/// <param name="cells">Uniform cells array.</param>
		public virtual void AddCells(string cellId, UniformDataArray cells)
		{
			if (this.cells.ContainsKey(cellId))
				throw new MveException("UnstrGrid already contains cell array " + cellId);

			if (!typeof(ICell).IsAssignableFrom(cells.ElemType))
				throw new MveException("UnstrGrid: Cell array have to contain subclasses of Visualization.ICell");

			this.cells.Add(cellId, cells);
			this.cellAttrs.Add(cellId, new Hashtable());
		}

		/// <summary>
		/// Remove array of cells and associated attributes.
		/// </summary>
		/// <param name="cellId">Unique (in curent UnstrGrid) ID of cell array.</param>
		public virtual void RemoveCells(string cellId)
		{
			if (!this.cells.ContainsKey(cellId))
				throw new MveException("UnstrGrid do not contain cell array " + cellId);

			this.cells.Remove(cellId);
			this.cellAttrs.Remove(cellId);
		}

		/// <summary>
		/// Returns cell array with given ID.
		/// </summary>
		/// <param name="cellId">Unique (in curent UnstrGrid) ID of cell array.</param>
		/// <returns>Cell array.</returns>
		public virtual UniformDataArray GetCells(string cellId)
		{
			if (!this.cells.ContainsKey(cellId))
				throw new MveException("UnstrGrid do not contain cell array " + cellId);

			return (UniformDataArray) this.cells[cellId];
		}    

		/// <summary>
		/// Returns array of identifiers of all cell arrays.
		/// </summary>
		/// <returns>Array of identifiers of all cell arrays.</returns>
		public virtual ICollection GetCellIds()
		{
			return this.cells.Keys;
		} 

		/// <summary>
		/// Add array of point attributes. 
		/// The length should be equal length of points array.
		/// Id have to be unique in frame of curent instance of UnstrGrid.
		/// </summary>
		/// <param name="attrId">Id of attribute array.</param>
		/// <param name="attrs">Uniform data array of point attributes.</param>
		public virtual void AddPointAttrs(string attrId, UniformDataArray attrs)
		{
			if (pointAttrs.ContainsKey(attrId))
			{
				throw new MveException("UnstrGrid already contains array of point attributes named " + attrId);
			}

			if (points != null && points.Count != attrs.Count)
				Trace.WriteLine("Count of point attributes do not agree with count of points.");

			pointAttrs.Add(attrId, attrs);
		}
    
		/// <summary>
		/// Returns collection of cell attribute array ids.
		/// </summary>
		/// <param name="cellId">ID of cell array.</param>
		/// <returns>Collection of cell attribute array ids.</returns>
		public ICollection GetCellAttrsIds(string cellId)
		{
			if (!this.cellAttrs.ContainsKey(cellId))
				throw new Exception("UnstrGrid do not contain cell array with ID " + cellId);

			return ((Hashtable) cellAttrs[cellId]).Keys;
		}

		/// <summary>
		/// Returns particular array of cell attributes.
		/// </summary>
		/// <param name="cellId">ID of cells.</param>
		/// <param name="attrId">ID of attributes.</param>
		/// <returns>Particular array of cell attributes. If name is not present, returns null.</returns>
		public UniformDataArray GetCellAttrs(string cellId, string attrId)
		{
			if (!this.cellAttrs.ContainsKey(cellId))
				throw new Exception("UnstrGrid do not contain cell array with ID " + cellId);

			Hashtable attrs = (Hashtable) cellAttrs[cellId];

			if (!attrs.ContainsKey(attrId))
				throw new Exception("Cells do not contain attribute array with ID " + attrId);

			return (UniformDataArray) attrs[attrId];
		}

		/// <summary>
		/// Returns collection of point attribute array.
		/// </summary>
		/// <returns>Collection of point attribute array.</returns>
		public ICollection GetPointAttrsIds()
		{
			return pointAttrs.Keys;
		}

		/// <summary>
		/// Returns particular array of point attributes.
		/// </summary>
		/// <param name="attrId">Name of attribute array.</param>
		/// <returns>Returns particular array of point attributes. If name is not present returns null.</returns>
		public UniformDataArray GetPointAttrs(string attrId)
		{
			if (!pointAttrs.ContainsKey(attrId))
				throw new Exception("UnstrGrid: Point attributes do not contain key " + attrId + ".");
        
			return (UniformDataArray) pointAttrs[attrId];
		}
    
		/// <summary>
		/// Get/Set an uniform array of points. Array have to be filled by childrens of Zcu.Mve.Visualization.Point.
		/// </summary>
		public virtual UniformDataArray Points
		{
			set
			{
				if (typeof(IPoint).IsAssignableFrom(value.ElemType))
				{
					this.points = value;
					this.boundingBox = new BoundingBox(this.points);
				}
				else
				{
					throw new MveException("Only childrens of Zcu.Mve.IPoint are allowed in UnstrGrid point array.");
				}
			}
			get
			{
				return points;
			}
		}

		/// <summary>
		/// Returns Min-Max box.
		/// </summary>
		public BoundingBox BoundingBox
		{
			get
			{
				return this.boundingBox;
			}
		}

		/// <summary>
		/// Returns whether cells named cellId are present.
		/// </summary>
		/// <param name="cellId">ID of cell.</param>
		/// <returns>True if cells named cellId are present, otherwise false.</returns>
		public bool ContainsCells(string cellId)
		{
			return this.cells.ContainsKey(cellId);
		}
 
		/// <summary>
		/// Returns whether point attributes named attrId are present.
		/// </summary>
		/// <param name="attrId">ID of point attributes.</param>
		/// <returns>True if point attributes named attrId are present, otherwise false.</returns>
		public bool ContainsPointAttrs(string attrId)
		{
			return this.pointAttrs.ContainsKey(attrId);
		}
   
		/// <summary>
		/// Returns whether cell attributes named attrId are present.
		/// </summary>
		/// <param name="cellId">ID of cell.</param>
		/// <param name="attrId">ID of attributes.</param>
		/// <returns>True if cell attributes named attrId are present, otherwise false.</returns>
		public bool ContainsCellAttrs(string cellId, string attrId)
		{
			if (!cellAttrs.ContainsKey(cellId))
				return false;

			Hashtable cellsAttrs = (Hashtable)cellAttrs[cellId];
      
			return cellsAttrs.ContainsKey(attrId);
		}

		/// <summary>
		/// Get/Set transform of curent object.
		/// </summary>
		public AbstractTransform Transform
		{
			get
			{
				return this.transform;
			}
			set
			{
				this.transform = value;
			}
		}

    
		/// <summary>
		/// Check consistence of curent structure and its substructures (arrays).
		/// </summary>
		/// <returns>If OK returns true, otherwise false.</returns>
		public override bool CheckConsistence()
		{
			bool isOK = true;
      
			// provereni bodu
			if (this.points == null)
			{
				Console.WriteLine("UnstrGrid do not contain points.");
				isOK = false;
			}
			else
			{
				if (this.points.CheckConsistence() == false)
					isOK = false;
			}
      
			// provereni bunek
			foreach (UniformDataArray c in this.cells.Values)
				if (c.CheckConsistence() == false)
					isOK = false;

			// provereni atributu bunek
			foreach (DictionaryEntry cell_entry in this.cellAttrs)
			{
				foreach (DictionaryEntry attr_entry in ((Hashtable) cell_entry.Value))
				{
					UniformDataArray array = (UniformDataArray) attr_entry.Value;

					if (array.CheckConsistence() == false)
						isOK = false;
          
					/*
					if (array.Count != cells.Count)
					{
					  Trace.WriteLine("Lenght of cell attributes array (" + entry.Key + ") do not agree with cell array count.");
					  isOK = false;
					}
					*/
				}
			}

			// provereni atributu bodu
			foreach (DictionaryEntry entry in this.pointAttrs)
			{
				UniformDataArray array = (UniformDataArray) entry.Value;

				if (array.CheckConsistence() == false)
					isOK = false;

				if (points != null && array.Count != points.Count)
				{
					Trace.WriteLine("Lenght of point attributes array (" + entry.Key + ") do not agree with point array count.");
					isOK = false;
				}

			}
			return isOK;
		} // CheckConsistence()

		/// <summary>
		/// Creates deep copy of current data structure (same data, different memory space). 
		/// Handle with care, UnstrGrid is ment to be large.
		/// </summary>
		/// <returns>Deep copy of current data structure.</returns>
		public override IDataObject DeepCopy()
		{
			UnstrGrid ret = new UnstrGrid();

			if (this.transform != null)
				ret.transform = (Transform3D) this.transform.DeepCopy();

			if (this.points != null)
				ret.points = (UniformDataArray) this.points.DeepCopy();

			ret.libNames = (ArrayList) this.libNames.Clone();

			foreach (DictionaryEntry entry in this.cells)
			{
				UniformDataArray array = (UniformDataArray) entry.Value;
				ret.AddCells(entry.Key.ToString(), (UniformDataArray) array.DeepCopy());
			}
      
			foreach (DictionaryEntry cellEntry in this.cellAttrs)
			{
				Hashtable table = (Hashtable) cellEntry.Value;

				foreach (DictionaryEntry attrEntry in table)
				{
					ret.AddCellAttrs(cellEntry.Key.ToString(), attrEntry.Key.ToString(), (UniformDataArray)((UniformDataArray) attrEntry.Value).DeepCopy());
				}
			}

			foreach (DictionaryEntry entry in this.pointAttrs)
			{
				ret.pointAttrs.Add(entry.Key, ((UniformDataArray) entry.Value).DeepCopy());
			}

			return ret;
		}
  
		/// <summary>
		/// Returns list of assembly names where can be found containg data types.
		/// </summary>
		public override string [] LibNames
		{
			get
			{
				base.libNames.Clear();

				AddLibName(typeof(UnstrGrid));

				if (points != null)
					AddLibName(points);

				foreach (UniformDataArray array in this.cells.Values)
				{
					AddLibName(array);
				}

				foreach (UniformDataArray attrs in this.pointAttrs.Values)
				{ 
					AddLibName(attrs);
				}   

				foreach (Hashtable table in this.cellAttrs.Values)
				{
					foreach (UniformDataArray attrs in table.Values)
					{ 
						AddLibName(attrs);
					}   
				}

				return base.LibNames;
			}
		}

  		/// <summary>
		/// Write current data to Xml file.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		public override void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			// Write transform
			if (this.transform != null)
			{
				xmlTextWriter.WriteStartElement("transform");
				xmlTextWriter.WriteAttributeString("type", this.transform.GetType().FullName);
				this.transform.WriteData(xmlTextWriter);
				xmlTextWriter.WriteEndElement();
			}

			// Write points
			if (points != null)
			{
				xmlTextWriter.WriteStartElement("points");
				points.WriteData(xmlTextWriter);
				xmlTextWriter.WriteEndElement();
			}

			// Write cells
			foreach (DictionaryEntry entry in this.cells)
			{
				UniformDataArray array = (UniformDataArray) entry.Value;
				xmlTextWriter.WriteStartElement("cells");
				xmlTextWriter.WriteAttributeString("cellid", entry.Key.ToString());
				array.WriteData(xmlTextWriter);
				xmlTextWriter.WriteEndElement();
			}
      
			// Write point attributes
			foreach (DictionaryEntry entry in this.pointAttrs)
			{
				UniformDataArray array = (UniformDataArray) entry.Value;
				xmlTextWriter.WriteStartElement("pointattrs");
				xmlTextWriter.WriteAttributeString("attrid", entry.Key.ToString());
				array.WriteData(xmlTextWriter);
				xmlTextWriter.WriteEndElement();
			}

			// Write cell attributes
			foreach (string cellId in this.GetCellIds())
			{
				foreach (string attrId in this.GetCellAttrsIds(cellId))
				{
					UniformDataArray array = (UniformDataArray) GetCellAttrs(cellId, attrId);
					xmlTextWriter.WriteStartElement("cellattrs");
					xmlTextWriter.WriteAttributeString("cellid", cellId);
					xmlTextWriter.WriteAttributeString("attrid", attrId);
					array.WriteData(xmlTextWriter);
					xmlTextWriter.WriteEndElement();
				}
			}
		}
  
		/// <summary>
		/// Read data from Xml file to current structure.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		public override void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			UniformDataArray array;
			string cellId;
			string attrId;

			while(xmlTextReader.Read())
			{
				if (xmlTextReader.NodeType == XmlNodeType.Element)
					switch(xmlTextReader.Name)
					{
						case "transform":
							this.transform = (AbstractTransform)Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
							this.transform.ReadData(xmlTextReader);
							break;
						case "points":		// naten bod
							array = new UniformDataArray();
							array.ReadData(xmlTextReader);
							this.Points = array;
							break;
						case "cells":		// naten bunk
							cellId = xmlTextReader.GetAttribute("cellid");
							array = new UniformDataArray();
							array.ReadData(xmlTextReader);
							AddCells(cellId, array);
							break;
						case "pointattrs":	// naten atribut bod
							attrId = xmlTextReader.GetAttribute("attrid");
							array = new UniformDataArray();
							array.ReadData(xmlTextReader);
							this.AddPointAttrs(attrId, array);
							break;
						case "cellattrs":	// naten atribut bunk
							cellId = xmlTextReader.GetAttribute("cellid");
							attrId = xmlTextReader.GetAttribute("attrid");
							array = new UniformDataArray();
							array.ReadData(xmlTextReader);
							this.AddCellAttrs(cellId, attrId, array);
							break;
						default:
							break;
					} // switch
			} // while
		}
  
		/// <summary>
		/// Returns a String that represents the current Object.
		/// </summary>
		/// <returns>String that represents the current Object.</returns>
		public override string ToString()
		{
			string str = this.GetType().FullName + ":\r\n";

			if (points != null)
			{
				// str += "********************************\r\n";
				str += "points: " + points.Count + "\r\n";
				str += "type of points: " + points.GetType().FullName + "\r\n";
			}

			if (cells != null)
			{
				str += "cell arrays: " + cells.Count + "\r\n";
				foreach (string cellId in this.GetCellIds())
				{
					UniformDataArray cellArray = this.GetCells(cellId);
					str += "    ID: " + cellId + " Type: " + cellArray.ElemType.FullName + " count: " + cellArray.Count +"\r\n";
				}
			}

			str += "point attributes: " + pointAttrs.Count + "\r\n";
			foreach (DictionaryEntry entry in pointAttrs)
			{
				UniformDataArray array = (UniformDataArray) entry.Value;
				str += "    ID: " + entry.Key + " Type: " + array.ElemType.FullName + "\r\n";
			}

			str += "cell attributes:\r\n";
			foreach (string cellId in this.GetCellIds())
			{
				str += "    CellID: " + cellId + "\r\n";
				foreach (string attrId in this.GetCellAttrsIds(cellId))
				{
					UniformDataArray array = (UniformDataArray) GetCellAttrs(cellId, attrId);
					str += "      AttrID: " + attrId + " Type: " + array.ElemType.FullName + "\r\n";
				}
			}
			return str;
		}

		/// <summary>
		/// Returns a shallow copy of the Unstructured Grid. 
		/// Useful when adding some attributes to an UnstrGrid you have 
		/// received as an input when you want to keep all the existing ones.
		/// </summary>
		/// <returns>A shallow copy of the original - has all the points, 
		/// cells and attributes of the original, but no copying is 
		/// actually performed.</returns>
		public UnstrGrid ShallowCopy()
		{
			UnstrGrid result = new UnstrGrid();

			// add points
			result.Points = this.Points;

			// add point attributes
			ICollection PointAttributeNames = this.GetPointAttrsIds();
			foreach(string id in PointAttributeNames)
			{
				result.AddPointAttrs(id,this.GetPointAttrs(id));
			}

			// add cells
			ICollection CellNames = this.GetCellIds();
			foreach(string id in CellNames)
			{
				result.AddCells(id, this.GetCells(id));

				// add cell attributes
				ICollection CellAttributeNames = this.GetCellAttrsIds(id);
				foreach (string attributeId in CellAttributeNames)
				{
					result.AddCellAttrs(id,attributeId,this.GetCellAttrs(id, attributeId));
				}
			}
			return(result);
		}

	} // class UnstrGrid
} // namespace
